module EncodingBtoH 

where
import Lexer
import Syntax
import Char
import Data.List
import Parser
import Decoding

-----------------------------------------------------------------------------------------------------------------------------------------------------
-- module EncodingBtoH performes translation of the calculus lambda(fcb) into the calculus lambda(fch) (\(fcb) -> \(fch))
-- the encoding of components buffer and empty buffer is realized during the semantic analysis with the function renameBoundedProcBufferToHandle,
-- and the encoding of constants buffer and get, and the expression put is realized during the semantic analysis with the function renameBoundedExp;
-- the function renameBoundedProcBufferToHandle is analogical to renameBoundedProc (module SemAnalysis) and performes semantic analysis as well,
-- i.e. it proves wether all the variables in processes are bounded to new operators (nu),
-- and renames all bounded variables using a list of fresh variables (freshVariables); 
-- for checking variable bindings in expressions the function renameBoundedExp  will be called;
-- the function renameBoundedExp is analogical to renameBoundedExp in the module SemAnalysis (please, see module SemAnalysis for more details)
-----------------------------------------------------------------------------------------------------------------------------------------------------
-- all encodings are realized in accordance with the equivalence of translation (\(fcb) -> \(fch)), described in Chapter 4 of our master's thesis
-----------------------------------------------------------------------------------------------------------------------------------------------------

freshVariables = ["newVar" ++show x|x<-[1..500]]  
            
renameBoundedProcBufferToHandle :: Proc -> Proc
renameBoundedProcBufferToHandle x  = fst (renameNewBufferToHandle x freshVariables [])

renameNewBufferToHandle :: Proc -> [Var] -> [(Var,Var)] ->(Proc, [Var])
renameNewBufferToHandle (Usedhandle a) y z = (Usedhandle (renameVar a z), y)
renameNewBufferToHandle (Thread a b) y z = let (b',y') = renameBoundedExp b y z in (Thread (renameVar a z) b', y')
renameNewBufferToHandle (Cell a b) y z = let (b',y') = renameBoundedExp b y z in (Cell (renameVar a z) b', y')
renameNewBufferToHandle (Lazythread a b) y z = let (b',y') = renameBoundedExp b y z in  (Lazythread (renameVar a z) b',y')
renameNewBufferToHandle (Handle a b) y z = (Handle (renameVar a z) (renameVar b z), y)
renameNewBufferToHandle (Buffer a b) y z = let s  = "<n>xp(<n>xg(<n>xs(<n>xh(" ++ a ++ "<= (xp,xg,xs,xh) | <n>h(<n>f (h <h> f | xp <c> f| xg <c> True| xs <c>" ++ show1 b ++ "|xh <c> h))))))"
                                        in (renameNewBufferToHandle (parser s) y z)
renameNewBufferToHandle (Emptybuffer a) y z = let s = "<n>xp(<n>xg(<n>xs(<n>xh(" ++ a ++ "<= (xp,xg,xs,xh)| <n>h(<n>f(<n>hs(<n>fs (h <h> f | hs <h> fs| xp <c> True| xg <c> f| xs <c> fs| xh <c> h))))))))"
                                     in (renameNewBufferToHandle (parser s) y z)
renameNewBufferToHandle (Parall p1 p2) y z = 
	                   let	(t1, t2) = (renameNewBufferToHandle p1 y z);
		                  (s1, s2) = (renameNewBufferToHandle p2 t2 z)
	                   in	((Parall t1 s1), s2)

renameNewBufferToHandle (New x p) (y:ys) mappings = let (new_p, freshVariables') = renameNewBufferToHandle p ys ((x,y):mappings)
                                                in (New y new_p, freshVariables') 

----------------------------------------------------------------------------------------------------------------------------------------------------------

renameVar :: Var -> [(Var, Var)] -> Var
renameVar x [] = error ("Variable " ++ x ++ "is  not bound") 
renameVar x ((z1,z2):zs) = if(z1 == x) then z2 else (renameVar x  zs)

----------------------------------------------------------------------------------------------------------------------------------------------------------

renameBoundedExp :: Exp -> [Var] -> [(Var, Var)] -> (Exp, [Var])
renameBoundedExp x fv mappings  = (renameLam x fv mappings) 

----------------------------------------------------------------------------------------------------------------------------------------------------------

renameLam :: Exp -> [Var] -> [(Var,Var)] ->(Exp, [Var])
renameLam (V x) y z = (V (renameVar x z), y)
renameLam (CCell) y z = (CCell, y)
renameLam (CThread) y z = (CThread, y)
renameLam (CHandle) y z = (CHandle, y)
renameLam (CLazy) y z = (CLazy, y)
renameLam (CUnit) y z = (CUnit, y)
renameLam (CBuffer) y z = renameLam bufferToHandle y z
renameLam (CGet) y z = renameLam getToHandle y z
renameLam (App x1 x2) y z = 
	let	(t1, t2) = (renameLam x1 y z);
		(s1, s2) = (renameLam x2 t2 z)
	in	((App t1 s1), s2)
renameLam (Exch x1 x2) y z = 
	let	(t1, t2) = (renameLam x1 y z);
		(s1, s2) = (renameLam x2 t2 z)
	in	((Exch t1 s1), s2)
renameLam (Put x1 x2) y z = renameLam (putToHandle x1 x2) y z  
renameLam (Lambda x1 x2) (y:ys) z =
		let	(t1, t2) = (renameLam x2 ys ((x1, y):z))
		in	((Lambda  y t1), t2)
renameLam (Constr x1 x2 x3 x4) y z =
		let	(t1, t2) = (renameInConstr z x4 y)
		in	((Constr x1 x2 x3 t1), t2)
renameLam (Case x1 x2) y z = 
		let	(t1, t2) = (renameLam x1 y z);
			(s1, s2) = (renameInCase z x2 t2)
		in	((Case t1 s1), s2)

--------------------------------------------------------------------------------------------------------------------------------------------------------

renameInConstr :: [(Var,Var)] -> [Exp] -> [Var]->([Exp], [Var])
renameInConstr x [] z = ([], z)
renameInConstr x (y:ys) z = 
            let	(t1, t2) = (renameLam y z x);  
                  (s1, s2) = (renameInConstr x ys t2)	in (t1:s1, s2)

--------------------------------------------------------------------------------------------------------------------------------------------------------

renameInCase :: [(Var,Var)]->[Alts]->[Var]->([Alts], [Var])
renameInCase x [] z = ([], z)
renameInCase x ((Alt y1 y2 0 [] y3):ys) z = 
		let	(t1, t2) = (renameLam y3 z x);
			(s1, s2) = (renameInCase x ys t2)
		in	((Alt y1 y2 0 [] t1):s1, s2)
renameInCase x ((Alt y1 y2 2 (v:w:[]) y3):ys) (z1:z2:zs) = 
		let	 (v1, w1, vnew, wnew) = (v, w, z1, z2); (t1, t2) = (renameLam y3 zs ((v1,z1):(w1,z2):x)); (s1, s2) = (renameInCase x ys t2)
		in	 ((Alt y1 y2 2 (vnew:wnew:[]) t1):s1, s2)
renameInCase x ((Alt y1 y2 3(u:v:w:[]) y3):ys) (z1:z2:z3:zs) = 
		let	 (u1, v1, w1, unew, vnew, wnew) = (u, v, w, z1, z2, z3); (t1, t2) = (renameLam y3 zs ((u1,z1):(v1,z2):(w1,z3):x)); (s1, s2) = (renameInCase x ys t2)
            in	 ((Alt y1 y2 3 (unew:vnew:wnew:[]) t1):s1, s2)
renameInCase x ((Alt y1 y2 4(q:u:v:w:[]) y3):ys) (z1:z2:z3:z4:zs) = 
		let	 (q1, u1, v1, w1, qnew, unew, vnew, wnew) = (q, u, v, w, z1, z2, z3, z4); (t1, t2) = (renameLam y3 zs ((q1,z1):(u1,z2):(v1,z3):(w1,z4):x)); (s1, s2) = (renameInCase x ys t2)
            in     ((Alt y1 y2 4 (qnew:unew:vnew:wnew:[]) t1):s1, s2)

---------------------------------------------------------------------------------------------------------------------------------------------------------
-- encoding of constants buffer and get, and expression put
---------------------------------------------------------------------------------------------------------------------------------------------------------


bufferToHandle = Lambda "new" (App (Lambda "new2" (Case (V "new2") [Alt 3 1 2 ["h","f"] (App (Lambda "new1" (Case (V "new1") [Alt 3 1 2 ["hs","fs"] (App (Lambda "putg" (App (Lambda "getg" (App (Lambda "stored" (App (Lambda "handler" (App CThread (Lambda "new" (Constr 5 1 4 [V "putg",V "getg",V "stored",V "handler"])))) (App CCell (V "h")))) (App CCell (V "fs")))) (App CCell (V "f")))) (App CCell (Constr 2 1 0 [])))])) (App CHandle (Lambda "fs" (Lambda "hs" (Constr 3 1 2 [V "hs",V"fs"])))))])) (App CHandle (Lambda "f" (Lambda "h" (Constr 3 1 2 [V "h",V "f"])))))

putToHandle x v = App (Lambda "new4" (Case (V "new4") [Alt 3 1 2 ["x","v"] (Case (V "x") [Alt 5 1 4 ["xp","xg","xs","xh"] (App (Lambda "new3" (Case (V "new3") [Alt 3 1 2 ["h","f"] (App (Lambda "new2" (App (Exch (V "xh") (V "h")) (Constr 2 1 0 []))) (App (Lambda "new1" (Exch (V "xs") v )) (Case (Exch (V "xp") (V "f")) [Alt 2 1 0 [] (Constr 2 1 0 []),Alt 2 2 0 [] (Constr 2 1 0 [])])))])) (App CHandle (Lambda "f" (Lambda "h" (Constr 3 1 2 [V "h",V "f"])))))])])) (Constr 3 1 2 [x, v])
 
getToHandle = Lambda "new5" (Case (V "new5") [Alt 5 1 4 ["xp","xg","xs","xh"] (App (Lambda "new4" (Case (V "new4") [Alt 3 1 2 ["h","f"] (App (Lambda "new3" (Case (V "new3") [Alt 3 1 2 ["hs","fs"] (App (Lambda "new2" (App (Lambda "v" (App (Lambda "new1" (V "v")) (App (Exch (V "xh") (V "h")) (Constr 2 1 0 [])))) (Exch (V "xs") (V "fs")))) (Case (Exch (V "xg") (V "f")) [Alt 2 1 0 [] (Constr 2 1 0 []),Alt 2 2 0 [] (Constr 2 1 0 [])]))])) (App CHandle (Lambda "fs" (Lambda "hs" (Constr 3 1 2 [V "hs",V "fs"])))))])) (App CHandle (Lambda "f" (Lambda "h" (Constr 3 1 2 [V "h",V "f"])))))])

---------------------------------------------------------------------------------------------------------------------------------------------------------
-- some tests
---------------------------------------------------------------------------------------------------------------------------------------------------------

test_h1 =  renameBoundedProcBufferToHandle . parser $ "<n> x (<n> y (y <= (\\w -> put(x,False) ) (get x) | x <b> True ))"
test_h2 =  renameBoundedProcBufferToHandle . parser $ "<n> x (<n> y (x <b> (\\w -> w) | y <eb> ))"


