module EncodingHtoB  

where
import Lexer
import Syntax
import Char
import Data.List
import Parser
import Debug.Trace
import Decoding

--------------------------------------------------------------------------------------------------------------------------------------------------
-- module EncodingHtoB performes translation of the calculus lambda(fch) into the calculus lambda(fcb) (\(fch) -> \(fcb))
-- the encoding of components handle and used handle is realized during the semantic analysis with the function renameBoundedProcHandeleToBuffer,
-- and the encoding of the constant handle is realized during the semantic analysis with the function renameBoundedExp;
-- the function renameBoundedProcHandeleToBuffer is analogical to renameBoundedProc (module SemAnalysis) and performes semantic analysis as well,
-- i.e. it proves wether all the variables in processes are bounded to new operators (nu),
-- and renames all bounded variables using a list of fresh variables (freshVariables); 
-- for checking variable bindings in expressions the function renameBoundedExp  will be called;
-- the function renameBoundedExp is analogical to renameBoundedExp in the module SemAnalysis (please, see module SemAnalysis for more details)
--------------------------------------------------------------------------------------------------------------------------------------------------
-- all encodings are realized in accordance with the equivalence of translation (\(fch) -> \(fcb)), described in Chapter 4 of our master's thesis
--------------------------------------------------------------------------------------------------------------------------------------------------


freshVariables = ["newVar" ++show x|x<-[1..500]]  

renameBoundedProcHandleToBuffer :: Proc -> Proc
renameBoundedProcHandleToBuffer x  = fst (renameNewHandleToBuffer x freshVariables [])

renameNewHandleToBuffer :: Proc -> [Var] -> [(Var,Var)] ->(Proc, [Var])
renameNewHandleToBuffer (Usedhandle a) y z = renameNewHandleToBuffer(Lazythread a (V a)) y z
renameNewHandleToBuffer (Thread a b) y z = let (b',y') = renameBoundedExp b y z in (Thread (renameVar a z) b', y')
renameNewHandleToBuffer (Cell a b) y z = let (b',y') = renameBoundedExp b y z in (Cell (renameVar a z) b', y')
renameNewHandleToBuffer (Lazythread a b) y z = let (b',y') = renameBoundedExp b y z in  (Lazythread (renameVar a z) b',y')
renameNewHandleToBuffer (Handle a b) (y1:y2:y3:y4:ys) z = let s = "<n>" ++ y1 ++ "(" ++ b ++ "<s= (\\" ++ y2 ++ "->((\\" ++ y3 ++ "-> put(" ++ y1 ++ "," ++ y2 ++ "))" ++y2 ++")) (get "++ y1 ++")|" ++  y1 ++ "<eb> |" ++ a ++ "<= \\" ++ y4 ++ "-> put (" ++ y1 ++"," ++y4++ "))"
                                            in (renameNewHandleToBuffer (parser s) ys z) 
renameNewHandleToBuffer (Buffer a b) y z = let (b',y') = renameBoundedExp b y z in (Buffer (renameVar a z) b',y')
renameNewHandleToBuffer (Emptybuffer a) y z = (Emptybuffer (renameVar a z), y)
renameNewHandleToBuffer (Parall p1 p2) y z = 
	                    let	(t1, t2) = (renameNewHandleToBuffer p1 y z);
	                   	(s1, s2) = (renameNewHandleToBuffer p2 t2 z)
	                    in	((Parall t1 s1), s2)

renameNewHandleToBuffer (New x p) (y:ys) mappings = let (new_p,freshVariables') = renameNewHandleToBuffer p ys ((x,y):mappings)
                                                in (New y new_p, freshVariables') 

---------------------------------------------------------------------------------------------------------------------------------------------------

renameVar :: Var -> [(Var, Var)] -> Var
renameVar x [] = error ("Variable " ++ x ++ " is  not bound") 
renameVar x ((z1,z2):zs) = if(z1 == x) then z2 else (renameVar x  zs)

---------------------------------------------------------------------------------------------------------------------------------------------------

renameBoundedExp :: Exp -> [Var] -> [(Var, Var)] -> (Exp, [Var])                             
renameBoundedExp x fv mappings  = (renameLam x fv mappings) 

----------------------------------------------------------------------------------------------------------------------------------------------------

renameLam :: Exp -> [Var] -> [(Var,Var)] ->(Exp, [Var])
renameLam (V x) y z = (V (renameVar x z), y)
renameLam (CCell) y z = (CCell, y)
renameLam (CThread) y z = (CThread, y)
renameLam (CHandle) (y1:y2:y3:y4:y5:y6:y7:y8:ys) z = renameLam (handleToBuffer y1 y2 y3 y4 y5 y6 y7 y8) ys z
renameLam (CLazy) y z = (CLazy, y)
renameLam (CUnit) y z = (CUnit, y)
renameLam (CBuffer) y z = (CBuffer, y)
renameLam (CGet) y z = (CGet, y)
renameLam (Lambda x1 x2) (y:ys) z =
		let	(t1, t2) = (renameLam x2 ys ((x1, y):z))
		in	((Lambda  y t1), t2)
renameLam (App x1 x2) y z = 
	let	(t1, t2) = (renameLam x1 y z);
		(s1, s2) = (renameLam x2 t2 z)
	in	((App t1 s1), s2)
renameLam (Exch x1 x2) y z = 
	let	(t1, t2) = (renameLam x1 y z);
		(s1, s2) = (renameLam x2 t2 z)
	in	((Exch t1 s1), s2)
renameLam (Put x1 x2) y z = 
	let	(t1, t2) = (renameLam x1 y z);
		(s1, s2) = (renameLam x2 t2 z)
	in	((Put t1 s1), s2)
renameLam (Constr x1 x2 x3 x4) y z =
		let	(t1, t2) = (renameInConstr z x4 y)
		in	((Constr x1 x2 x3 t1), t2)
renameLam (Case x1 x2) y z = 
		let	(t1, t2) = (renameLam x1 y z);
			(s1, s2) = (renameInCase z x2 t2)
		in	((Case t1 s1), s2)

----------------------------------------------------------------------------------------------------------------------------------------------------

renameInConstr :: [(Var,Var)] -> [Exp] -> [Var]->([Exp], [Var])
renameInConstr x [] z = ([], z)
renameInConstr x (y:ys) z = 
            let	(t1, t2) = (renameLam y z x);  
                  (s1, s2) = (renameInConstr x ys t2)	in (t1:s1, s2)

----------------------------------------------------------------------------------------------------------------------------------------------------

renameInCase :: [(Var,Var)]->[Alts]->[Var]->([Alts], [Var])
renameInCase x [] z = ([], z)
renameInCase x ((Alt y1 y2 0 [] y3):ys) z = 
		let	(t1, t2) = (renameLam y3 z x);
			(s1, s2) = (renameInCase x ys t2)
		in	((Alt y1 y2 0 [] t1):s1, s2)
renameInCase x ((Alt y1 y2 2 (v:w:[]) y3):ys) (z1:z2:zs) = 
		let	 (v1, w1, vnew, wnew) = (v, w, z1, z2); (t1, t2) = (renameLam y3 zs ((v1,z1):(w1,z2):x)); (s1, s2) = (renameInCase x ys t2)
		in	 ((Alt y1 y2 2 (vnew:wnew:[]) t1):s1, s2)
renameInCase x ((Alt y1 y2 3(u:v:w:[]) y3):ys) (z1:z2:z3:zs) = 
		let	 (u1, v1, w1, unew, vnew, wnew) = (u, v, w, z1, z2, z3); (t1, t2) = (renameLam y3 zs ((u1,z1):(v1,z2):(w1,z3):x)); (s1, s2) = (renameInCase x ys t2)
            in	 ((Alt y1 y2 3 (unew:vnew:wnew:[]) t1):s1, s2)
renameInCase x ((Alt y1 y2 4(q:u:v:w:[]) y3):ys) (z1:z2:z3:z4:zs) = 
		let	 (q1, u1, v1, w1, qnew, unew, vnew, wnew) = (q, u, v, w, z1, z2, z3, z4); (t1, t2) = (renameLam y3 zs ((q1,z1):(u1,z2):(v1,z3):(w1,z4):x)); (s1, s2) = (renameInCase x ys t2)
            in     ((Alt y1 y2 4 (qnew:unew:vnew:wnew:[]) t1):s1, s2)

----------------------------------------------------------------------------------------------------------------------------------------------------------
-- encoding of the constant handle 
----------------------------------------------------------------------------------------------------------------------------------------------------------

handleToBuffer y1 y2 y3 y4 y5 y6 y7 y8 = Lambda y1 (App (Lambda y2 (App (Lambda y3 (App (Lambda y4 (App (App (V y1) (V y3)) (V y4))) (App CThread (Lambda y5 (Lambda y6 (Put (V y2) (V y6))))))) (App CLazy (Lambda y5 (App (Lambda y7 (App (Lambda y8 (V y7)) (Put (V y2) (V y7)))) (App CGet (V y2))))))) (App CBuffer CUnit))

----------------------------------------------------------------------------------------------------------------------------------------------------------
-- some tests
----------------------------------------------------------------------------------------------------------------------------------------------------------

test_b1 = renameBoundedProcHandleToBuffer . parser $ "<n> x (x <= handle (\\x -> \\y -> (y True)))"
test_b2 = renameBoundedProcHandleToBuffer . parser $ "<n> x (<n> y (x <h> y ))"
test_b3 = renameBoundedProcHandleToBuffer . parser $ "<n> x (x <_> )"

