module Evaluation (mayConvergent, allEndPoints, totalMustConvergent, isSuccessful, )

where
import Parser
import SemAnalysis
import Syntax
import Transformation
import Char
import Data.List
import Maybe
import Debug.Trace
import EncodingBtoH
import Decoding
import ReductionRules
 
-------------------------------------------------------------------------------------------------------------------------------------------------------
-- the module Evaluation is 



frv = ["frVar" ++show x|x<-[1..500]]


-------------------------------------------------------------------------------------------------------------------------------------------------------
-- the function isIrreducible tests wether a process is irreducible, i.e. evaluation of all components returns Nothing
-------------------------------------------------------------------------------------------------------------------------------------------------------

isIrreducible :: Proc' -> Bool
isIrreducible ps = all isNothing (all1StepSuccessors ps)


-------------------------------------------------------------------------------------------------------------------------------------------------------
-- the function isSuccessful tests (involving functions isValue, boundToSP and findSuccessor) wether a process is successful, 
-- i.e. when all its components are successful:
-- Threads are successful if the expression is a value (function isValue),
-- and the variable-case (functions boundToSP) is already covered by the previous case (function findSuccessor), e.g. x <= y | y <= z | z <= True;
-- Cells, Buffers, Handles, and Suspended threads are always successful (componentsSuccessful _ ps)
-------------------------------------------------------------------------------------------------------------------------------------------------------

isSuccessful :: Proc' -> Bool
isSuccessful (Proc' ns ps) =  all (==True) (map (\p -> componentSuccessful p ps) ps)

-------------------------------------------------------------------------------------------------------------------------------------------------------
-- componentSuccessful tests wether components are successful:
-- Threads are successful if the expression is a value (function isValue),
-- and the variable-case (functions boundToSP) is already covered by the previous case (function findSuccessor), e.g. x <= y | y <= z | z <= True;
-- Cells, Buffers, Handles, and Lazy Threads are always successful (componentsSuccessful _ ps)
-------------------------------------------------------------------------------------------------------------------------------------------------------

componentSuccessful :: Comp -> [Comp] -> Bool
componentSuccessful (Thread' v (V x)) ps = boundToSP x [x] ps
componentSuccessful (Thread' v e) ps = isValue e                                                           
componentSuccessful _ ps = True                      

--------------------------------------------------------------------------------------------------------------------------------------------------------
-- boundToSP together with findSuccessor test wether all identifiers in threads are bound via a chain x_1 <= x_2 |..| x_n <= v to a non-variable values,
-- Cells, Handles, Used Handles, Lazy Threads , Buffers or Empty Buffers;
-- boundToSP returns False if a cyclic chain occurs, e.g. x <= y | y <= z | z <= x
--------------------------------------------------------------------------------------------------------------------------------------------------------

boundToSP :: Var -> [Var] -> [Comp] -> Bool
boundToSP x visitedvars ps =
                       let successor = findSuccessor x ps
                       in  case successor of
                                (Thread' v (V y)) -> if  y `elem` visitedvars then False else boundToSP y (x:y:visitedvars) ps
                                (Thread' v e)     -> isValue e
                                other             -> True

findSuccessor :: Var -> [Comp] -> Comp
findSuccessor x (p:ps) =
  let v = case p of
            (Thread' y e) -> [y]
            (Cell' y e)   -> [y]
            (Handle' y1 y2) -> [y1,y2]
            (Usedhandle' y) -> [y]
            (Buffer' y e) -> [y]
            (Emptybuffer' y) -> [y]
            (Lazythread' y e) -> [y]
  in if x `elem` v then p else findSuccessor x ps

-------------------------------------------------------------------------------------------------------------------------------------------------------
-- allEndpoints computes all irreducible successors wrt. -*,ev->
-- it returns a list of all possible endpoints of a process evaluation
-------------------------------------------------------------------------------------------------------------------------------------------------------

allEndPoints :: Proc' -> [Proc']
allEndPoints p = nub $ map fst $ allEP_it [(p,frv)]

allEP_it :: [(Proc', [Var])] -> [(Proc', [Var])]
allEP_it [] = []
allEP_it p =
 let xs  =  filter (\(p',v) -> isIrreducible p') p
     ys  =  concatMap (map fromJust . filter isJust . all1StepSuccessorsWV) (filter (\(p',v) -> not $ isIrreducible p') p) in xs ++ (allEP_it ys)


------------------------------------------------------------------------------------------------------------------------------------------------------
-- the function alliStepSucc performes i-step evaluation of all irreducible successors
------------------------------------------------------------------------------------------------------------------------------------------------------ 

alliStepSucc :: (Num t) => t -> Proc' -> [Proc']
alliStepSucc i p =  nub $ map fst $ allEP_it_i i [(p,frv)]

allEP_it_i :: (Num t) => t -> [(Proc', [Var])] -> [(Proc', [Var])]
allEP_it_i 0 p = p
allEP_it_i i [] = []
allEP_it_i i p =
 let xs  =  filter (\(p',v) -> isIrreducible p') p
     ys  =  concatMap (map fromJust . filter isJust . all1StepSuccessorsWV) (filter (\(p',v) -> not $ isIrreducible p') p) in xs ++ (allEP_it_i (i-1) ys)

--------------------------------------------------------------------------------------------------------------------------------------------------------
-- the function mayConvergent tests wether a process is may-convergent, i.e. that there is at least one sequence of successful reductions of a process
--------------------------------------------------------------------------------------------------------------------------------------------------------

mayConvergent :: Proc' -> Bool
mayConvergent p =  any (== True) [any (isSuccessful) (alliStepSucc i p) | i <- [1..]]

--------------------------------------------------------------------------------------------------------------------------------------------------------
-- the function totalMustConvergent tests wether a process is total must-convergent, i.e. all 
--------------------------------------------------------------------------------------------------------------------------------------------------------

totalMustConvergent :: Proc' -> Bool
totalMustConvergent p = all isSuccessful (allEndPoints p)

-------------------------------------------------------------------------------------------------------------------------------------------------------
-- the function all1StepSuccessors executes a one-step evaluation of all components of a process, 
-- whereas the function eval1Step will be invoked for fair evaluation of all eager threads;
-- the function returns a list of evaluated components
-------------------------------------------------------------------------------------------------------------------------------------------------------

all1StepSuccessors :: Proc' -> [Maybe (Proc', [Var])]
all1StepSuccessors (Proc' ns ps) = map eval1Step ( map (Proc' ns) (everyThreadFirst [] ps))
 where
  everyThreadFirst _ [] = []
  everyThreadFirst qs (p:ps) = ((p:ps) ++ qs) : everyThreadFirst (qs++[p]) ps 

-------------------------------------------------------------------------------------------------------------------------------------------------------
-- the function eval1Step performs a one-step evaluation of the first component in a process,
-- and calls the function eval1 (module ReductionRules) with a list of fresh variables (frv);
-- the function returns either Nothing, if the first component is not an eager thread, or an evaluated eager thread 
-------------------------------------------------------------------------------------------------------------------------------------------------------
 
eval1Step :: Proc' -> Maybe (Proc', [Var])
eval1Step (Proc' ns (p:ps)) = eval1 p ns ps frv


--------------------------------------------------------------------------------------------------------------------------------------------------------
-- the functions all1StepSuccessorsWV and eval1StepWV are analogical to all1StepSuccessorsWV and eval1StepWV (see above),
-- but have a list of fresh variable as additional argument
--------------------------------------------------------------------------------------------------------------------------------------------------------

all1StepSuccessorsWV :: (Proc', [Var]) -> [Maybe (Proc', [Var])]
all1StepSuccessorsWV ((Proc' ns ps),fv) = map (eval1StepWV fv) ( map (Proc' ns) (everyThreadFirst [] ps))
 where
  everyThreadFirst _ [] = []
  everyThreadFirst qs (p:ps) = ((p:ps) ++ qs) : everyThreadFirst (qs++[p]) ps

eval1StepWV :: [Var] -> Proc' -> Maybe (Proc', [Var])
eval1StepWV frv (Proc' ns (p:ps))  = eval1 p ns ps frv

-------------------------------------------------------------------------------------------------------------------------------------------------------
-- some tests
-------------------------------------------------------------------------------------------------------------------------------------------------------

run1Step = all1StepSuccessors . transfToProc' . parser
toProc' = transfToProc' . parser
toProcFCB' = transfToProcHtoB' . parser
toProcFCH' = transfToProcBtoH' . parser

test_endpoints = putStrLn $ unlines $ map show $ allEndPoints  $  toProc' $ "<n> x (<n> a (<n> b (<n> c (x <c> True | a <= c a | b <= c b | c <= \\z ->(thread (\\y -> exch(x,z)))))))"         

test_e1 = run1Step  "<n> x ( <n> y ((x <= y x | y <= \\ z -> z)))"
test_e2 = run1Step  "<n> x ( <n> y (<n> w (<n> b (x <= ((w y) y) | y <= w | w <= \\ z -> z | b <= ((w y)y )))))"
test_e3 = mayConvergent $ toProc' "<n> x (<n> y (y <= (\\w -> put(x,False) ) (get x) | x <b> True ))"
test_e4 = mayConvergent $ toProcFCH' "<n> x (<n> y (y <= (\\w -> put(x,False) ) (get x) | x <b> True ))"
test_e5 = totalMustConvergent $ toProc' "<n> x (<n> y (y <= (\\w -> put(x,False) ) (get x) | x <b> True ))"
test_e6 = totalMustConvergent $ toProcFCH' "<n> x (<n> y (y <= (\\w -> put(x,False) ) (get x) | x <b> True ))"
test_e7 = mayConvergent $ toProc' "<n> x (x <= handle (\\x -> \\y -> ((\\w -> (x True))(y (\\q -> q)))))"
test_e8 = mayConvergent $ toProcFCH' "<n> x (x <= handle (\\x -> \\y -> ((\\w -> (x True))(y (\\q -> q)))))"
test_e9 = totalMustConvergent $ toProc' "<n> x (x <= handle (\\x -> \\y -> ((\\w -> (x True))(y (\\q -> q)))))"
test_e10 = totalMustConvergent $ toProcFCH' "<n> x (x <= handle (\\x -> \\y -> ((\\w -> (x True))(y (\\q -> q)))))"


