module Lexer (Token(..), Position, getPos, printToken, lexer)
 where
 import Data.Char
 import Data.List
 import Syntax

-----------------------------------------------------------------------------------------------------------------------
-- module Lexer performs the lexical analysis of a source program
-- data Token is a data structure of tokens that may be used in a source program 
-- the argument Position has type (Int,Int)=(line, column) and determines the position of a token in a source program   
-----------------------------------------------------------------------------------------------------------------------

 data Token = TokInt	     Position   Integer		      -- integer
	      | TokVar	     Position   String			-- identifier	
 	      | TokTrue	     Position		            -- True
	      | TokFalse       Position				-- False
	      | TokNil	     Position				-- []
            | TokCons	     Position				-- :
            | TokCCell       Position                       -- cell
            | TokCThread     Position                       -- thread 
            | TokCHandle     Position                       -- handle
            | TokCLazy       Position                       -- lazy  
            | TokCUnit       Position                       -- unit
            | TokCBuffer     Position                       -- buffer 
            | TokCGet        Position                       -- get
            | TokExch        Position                       -- exch
            | TokPut         Position                       -- put 
	      | TokCase        Position                       -- case
	      | TokOf	     Position				-- of
	      | TokComma	     Position				-- ,
		| TokCBOpen	     Position				-- {
		| TokCBClose     Position				-- }
		| TokArrow	     Position				-- ->
		| TokLam	     Position				-- \\ (lambda)
		| TokBOpen	     Position				-- (
		| TokBClose	     Position				-- )
		| TokEq	     Position				-- =
            | TokThread	     Position				-- <=  
            | TokSusThread   Position				-- <s=
            | TokNew         Position				-- <n>
            | TokCell        Position				-- <c>
            | TokHandle      Position				-- <h>
            | TokUsedHandle  Position			      -- <_>
            | TokBuffer      Position				-- <b>
            | TokEmptyBuffer Position			      -- <eb>
            | TokParall      Position				-- |
            | Error          Position
			deriving (Eq,Show)
 
 type Position = (Int, Int)

------------------------------------------------------------------------------------------------------
-- function getPos returns the position of a token as two integers (Int,Int) = (line, column) 
------------------------------------------------------------------------------------------------------

 getPos :: Token -> Position
 getPos (TokInt x y) = x
 getPos (TokVar x y) = x
 getPos (TokTrue x)=x
 getPos (TokFalse x)=x
 getPos (TokNil x)=x
 getPos (TokCons x)=x
 getPos (TokCCell x)=x
 getPos (TokCThread x)=x
 getPos (TokCLazy x)=x
 getPos (TokCUnit x)=x
 getPos (TokCHandle x)=x
 getPos (TokCBuffer x)=x
 getPos (TokCGet x)=x
 getPos (TokPut x)=x
 getPos (TokExch x)=x
 getPos (TokCase x)=x
 getPos (TokOf x)=x
 getPos (TokComma x)=x
 getPos (TokCBOpen x)=x
 getPos (TokCBClose x)=x
 getPos (TokArrow x)=x
 getPos (TokLam x)=x
 getPos (TokBOpen x)=x
 getPos (TokBClose x)=x
 getPos (TokEq x)=x
 getPos (TokThread x) = x
 getPos (TokSusThread x) = x
 getPos (TokNew x) = x
 getPos (TokCell x) = x
 getPos (TokHandle x) = x
 getPos (TokUsedHandle x) = x
 getPos (TokBuffer x) = x
 getPos (TokEmptyBuffer x) = x
 getPos (TokParall x) = x
 getPos (Error x) = x
 
-------------------------------------------------------------------------------------------------
-- function printToken returns an original view of a token, so as it is used in a source program
-------------------------------------------------------------------------------------------------

 printToken :: Token -> String
 printToken (TokInt x y) = show y
 printToken (TokVar x y) = y
 printToken (TokTrue _) = "True"
 printToken (TokFalse _) = "False"
 printToken (TokNil _) = "[]"
 printToken (TokCons _) = ":"
 printToken (TokCCell _) = "cell"
 printToken (TokCThread _) = "thread"
 printToken (TokCLazy _) = "lazy"
 printToken (TokCUnit _) = "unit"
 printToken (TokCHandle _) = "handle"
 printToken (TokCBuffer _) = "buffer"
 printToken (TokCGet _) = "get"
 printToken (TokPut _) = "put"
 printToken (TokExch _) = "exch"
 printToken (TokCase _) = "case"
 printToken (TokOf _) = "of"
 printToken (TokComma _) = ","
 printToken (TokCBOpen _) = "{"
 printToken (TokCBClose _) = "}"
 printToken (TokArrow _) = "->"
 printToken (TokLam _) = "/"
 printToken (TokBOpen _) = "("
 printToken (TokBClose _) = ")"
 printToken (TokEq _) = "="
 printToken (TokThread _) = "<="
 printToken (TokSusThread _) = "<s="
 printToken (TokNew _) = "<n>"
 printToken (TokCell _) = "<c>"
 printToken (TokHandle _) = "<h>"
 printToken (TokUsedHandle _) = "<_>"
 printToken (TokBuffer _) = "<b>"
 printToken (TokEmptyBuffer _) = "<eb>"
 printToken (TokParall _) = "|"
 printToken (Error _ ) = ""  

----------------------------------------------------------------------------------------------------------------------------------------
-- lexer is the main function in this module
-- for example, the function call is :  lexer "<n> a ( <n> b ( <n> a1 (a <= 15 | a1 <h> b)))"
-- it receives a source program as string and returns a stream of tokens with their positions of type [Token]
-- lexer1 is a subfunction of the function lexer, it starts with the position (1,1) and carries pattern matching for each character out 
---------------------------------------------------------------------------------------------------------------------------------------- 
 
 lexer :: String -> [Token]
 lexer str = lexer1 1 1 str

 lexer1 l c []=[]
 lexer1 l c ('\n':xs)= lexer1 (l+1) 0 xs                                                           -- pattern matching of new line
 lexer1 l c ('\t':xs)= lexer1 l (c+1) xs						                           -- pattern matching of tabulator   
 lexer1 l c ('\r':xs)= lexer1 (l+1) 0 xs                                                           -- pattern matching of new line   
 lexer1 l c ('[':(']':xs))= TokNil (l,c) : lexer1 l(c+2) xs
 lexer1 l c (':':xs)= TokCons (l,c) : lexer1 l(c+1) xs
 lexer1 l c (',':xs)= TokComma (l,c) : lexer1 l(c+1) xs
 lexer1 l c ('{':xs)= TokCBOpen (l,c) : lexer1 l(c+1) xs
 lexer1 l c ('}':xs)= TokCBClose (l,c) : lexer1 l(c+1) xs
 lexer1 l c ('-':('>':xs))= TokArrow (l,c) : lexer1 l(c+2) xs
 lexer1 l c ('<':('=':xs))= TokThread (l,c) : lexer1 l(c+2) xs
 lexer1 l c ('<':('s':('=':xs)))= TokSusThread (l,c) : lexer1 l(c+3) xs
 lexer1 l c ('<':('n':('>':xs)))= TokNew (l,c) : lexer1 l(c+3) xs
 lexer1 l c ('<':('c':('>':xs)))= TokCell (l,c) : lexer1 l(c+3) xs
 lexer1 l c ('<':('h':('>':xs)))= TokHandle (l,c) : lexer1 l(c+3) xs
 lexer1 l c ('<':('_':('>':xs)))= TokUsedHandle (l,c) : lexer1 l(c+3) xs
 lexer1 l c ('<':('b':('>':xs)))= TokBuffer (l,c) : lexer1 l(c+3) xs
 lexer1 l c ('<':('e':('b':('>':xs))))= TokEmptyBuffer (l,c) : lexer1 l(c+4) xs
 lexer1 l c ('|':xs)= TokParall (l,c) : lexer1 l(c+1) xs
 lexer1 l c ('(':xs)= TokBOpen (l,c) : lexer1 l(c+1) xs
 lexer1 l c (')':xs)= TokBClose (l,c) : lexer1 l(c+1) xs
 lexer1 l c ('=':xs)= TokEq (l,c) : lexer1 l(c+1) xs
 lexer1 l c ('\\':xs)= TokLam (l,c) : lexer1 l(c+1) xs
 lexer1 l c ('-':('-':xs))= lexer1 (l+1) 0 (noComments xs)
 
 lexer1 l c ('-': xs) = TokInt (l,c) (negate(read num)) : lexer1 l(c+ length (num)+1) rest         -- pattern matching of negative integers
     where (num,rest) = span isDigit xs
 	
 lexer1 l c (x:xs) 
     | isSpace x = lexer1 l(c+1) xs                                                                -- pattern matching of space  
     | isAlpha x = lexVar l c (x:xs)
     | isDigit x = lexNum l c (x:xs) 
     | otherwise = error(errorPrint (Error (l,c))++ show x) 

 lexNum l c xs = TokInt (l,c)(read num) : lexer1 l(c+ length (num)) rest                           -- pattern matching integers
     where (num,rest) = span isDigit xs

 lexVar l c xs =                                                                                   -- pattern matching of key words
   case span isAlphaDigit xs of
      ("False",rest)   -> TokFalse (l,c) : lexer1 l(c+5) rest 																		
      ("True",rest)    -> TokTrue (l,c) : lexer1 l(c+4) rest
      ("cell",rest)    -> TokCCell (l,c) : lexer1 l(c+4) rest
      ("thread",rest)  -> TokCThread (l,c) : lexer1 l(c+6) rest
      ("lazy",rest)    -> TokCLazy (l,c) : lexer1 l(c+4) rest
      ("unit",rest)    -> TokCUnit (l,c) : lexer1 l(c+4) rest
      ("exch",rest)    -> TokExch (l,c) : lexer1 l(c+4) rest
      ("case",rest)    -> TokCase (l,c) : lexer1 l(c+4) rest
      ("of",rest)      -> TokOf (l,c) : lexer1 l(c+2) rest
      ("buffer",rest)  -> TokCBuffer (l,c) : lexer1 l(c+6) rest
      ("get",rest)     -> TokCGet (l,c) : lexer1 l(c+3) rest
      ("put",rest)     -> TokPut (l,c) : lexer1 l(c+3) rest
      ("handle",rest)  -> TokCHandle (l,c) : lexer1 l(c+6) rest
      (var,rest)       -> TokVar (l,c) var : lexer1 l(c+ length(var))  rest                        -- pattern matching of other identifiers
   where isAlphaDigit x = isAlpha x || isDigit x                                                   -- identifiers may be alphanumeric

---------------------------------------------------------------------------------------------------------------------------------------------
-- noComments deletes all comments beginnig with --
---------------------------------------------------------------------------------------------------------------------------------------------

 noComments []=[]										
 noComments (x:xs)= if (x=='\n') then xs else noComments xs

---------------------------------------------------------------------------------------------------------------------------------------------
-- errorPrint reports about characters that may not occur in a source program, i.e. characters that are not defined in the data Token
---------------------------------------------------------------------------------------------------------------------------------------------
 
 errorPrint x =  "\n *** Error during lexing at the position " ++ show(getPos x) ++ "\n unknown character: " 		 

---------------------------------------------------------------------------------------------------------------------------------------------
-- some tests
---------------------------------------------------------------------------------------------------------------------------------------------

