{module Parser (parser, parse)
where
import Lexer
import Syntax
import Data.List

------------------------------------------------------------------------------------------------------------------------------------------------
-- module Parser performs the syntax analysis of a source program
-- the function parser is the core function in this module, its receives a source code as a string and returns a code of type Proc,
-- the function call is, for example: parser "<n> a ( <n> b ( <n> a1 (a <= \\ x -> x | a1 <h> b)))"
-- the file Parser.y contains a parser specification for further parser generation with Happy, the Haskell Parser Generator
-- the parser has the name parse and handles tokens of type Token
-- we specify here the context-free grammar of the calculu lambda(fchb) in accordance with dats structures described in the module Syntax
-- and some actions for its syntax 
------------------------------------------------------------------------------------------------------------------------------------------------
}

%name parse
%tokentype { Token }
%token 
  int       { TokInt _ $$ }
  var	      { TokVar _ $$ }
  'False'   { TokFalse _  }
  'True'    { TokTrue _   }
  '[]'      { TokNil _    }
  ':'       { TokCons _   }
  'case'    { TokCase _   }
  'of'      { TokOf _     }
  'cell'    { TokCCell _   }
  'thread'  { TokCThread _   }
  'handle'  { TokCHandle _   }
  'lazy'    { TokCLazy _   }
  'unit'    { TokCUnit _   }
  'buffer'  { TokCBuffer _   }
  'get'     { TokCGet _   }
  'exch'    { TokExch _   }
  'put'     { TokPut _   }
  '<='      { TokThread _   }
  '<s='     { TokSusThread _   }
  '<n>'     { TokNew _   }
  '<c>'     { TokCell _   }
  '<h>'     { TokHandle _   }
  '<b>'     { TokBuffer _   }
  '<eb>'    { TokEmptyBuffer _   }
  '<_>'     { TokUsedHandle _   }
  '|'       { TokParall _   }
  ','       { TokComma _  }
  '{'       { TokCBOpen _ }
  '}'       { TokCBClose _ }
  '('       { TokBOpen _   }
  ')'       { TokBClose _  }
  '->'      { TokArrow _   }
  '\\'      { TokLam _     }
  '='       { TokEq _      }
  ''        { Error _    }   

%right '->'
%right':'
%left '|'
%nonassoc ','

%%
Proc :: {Proc}
Proc  : Comp '|' Proc                                { Parall $1 $3 }
      | Comp                                         { $1 }
      | '(' Proc ')'                                 { $2 }

Comp  : '<n>' var '(' Proc ')'                       { New (mkV $2) $4}                                        
      | var '<c>' Exp                                { Cell (mkV $1) (chkVal $2 $3) }
      | var '<='  Exp                                { Thread (mkV $1) $3 }
      | var '<h>' var                                { Handle (mkV $1) (mkV $3) }
      | var '<_>'                                    { Usedhandle (mkV $1) } 
      | var '<s=' Exp                                { Lazythread (mkV $1) $3 }
      | var '<b>' Exp                                { Buffer (mkV $1) (chkVal $2 $3) }
      | var '<eb>'                                   { Emptybuffer (mkV $1) } 

Exp : Exp AExp                                       { App $1 $2 }
    | AExp                                           { $1 }

AExp: var                                            { V (mkV $1)}
    | int                                            { mkInt $1 }       
    | '\\' var '->' AExp                             { Lambda (mkV $2) $4 }
    | 'exch' '(' AExp ',' AExp ')'                   { Exch  $3 $5 }
    | 'put' '(' AExp ',' AExp ')'                    { Put  $3 $5 }
    | 'case' AExp 'of' '{' Alts '}'                  { Case $2 (chkAlts $5) }  
    | '(' AExp ',' AExp ')'                          { pair  $2 $4 }
    | '(' AExp ',' AExp ',' AExp ')'                 { tupel3 $2 $4 $6}
    | '(' AExp ',' AExp ',' AExp ',' AExp ')'        { tupel4 $2 $4 $6 $8}
    | '(' AExp ':' AExp ')'                          { cons $2 $4 }                     
    | 'False'                                        { false }
    | 'True'                                         { true }
    | '[]'                                           { nil }
    | Constant                                       { $1 }
    | '(' Exp ')'                                    { $2 }
    
Constant : 'cell'                                    { CCell }
         | 'thread'                                  { CThread } 
         | 'handle'                                  { CHandle }
         | 'lazy'                                    { CLazy }
         | 'unit'                                    { CUnit }  
         | 'buffer'                                  { CBuffer }  
         | 'get'                                     { CGet }  

Alts : Alt                                           { [$1] }
     | Alt ',' Alts                                  { $1 : $3 }

Alt : Pat '->' Exp                                   { $1 $3 }

Pat  : 'True'                                        { altTrue }
     | 'False'                                       { altFalse }
     | '(' Pat ')'                                   { $2 }
     | var ':' var                                   { chkAltCons (mkV $1) (mkV $3)}
     | '[]'                                          { altNil }
     | '(' var ',' var ')'                           { chkAltPair (mkV $2) (mkV $4)}
     | '(' var ',' var ',' var ')'                   { chkAlt3Tupel (mkV $2) (mkV $4) (mkV $6)}
     | '(' var ',' var ',' var',' var ')'            { chkAlt4Tupel (mkV $2) (mkV $4) (mkV $6) (mkV $8)}

{
---------------------------------------------------------------------------------
-- happyError reports when the syntax structure of a source program is incorrect
---------------------------------------------------------------------------------

happyError :: [Token] -> a
happyError xs = error $ "parse error!" ++ show xs

----------------------------------------------------------------------------------
-- now we define actions
----------------------------------------------------------------------------------
-- mkV is an id function for identifiers
----------------------------------------------------------------------------------

mkV = id

---------------------------------------------------------------------------------------------------------------------------------------------
-- all constructors and alternatives have the form Constr Int Int Int [Exp] and Alt Int Int Int [Var] Exp, respectevely
-- these integer parameters are : type, number of a construct or an alternative of the same type, and arity
---------------------------------------------------------------------------------------------------------------------------------------------

nil             = Constr 1 1 0 []
cons a b        = Constr 1 2 2 [a,b]
true            = Constr 2 1 0 []
false           = Constr 2 2 0 []
pair a b        = Constr 3 1 2 [a,b]
tupel3 a b c    = Constr 4 1 3 [a,b,c]
tupel4 a b c d  = Constr 5 1 4 [a,b,c,d]
zero            = Constr 6 1 0 []
one             = Constr 6 2 0 []
npair a b       = Constr 6 3 2 [a,b]


altNil             = Alt 1 1 0 []
altCons a b        = Alt 1 2 2 [a,b]
altTrue            = Alt 2 1 0 []
altFalse           = Alt 2 2 0 []
altPair a b        = Alt 3 1 2 [a,b]
alt3Tupel a b c    = Alt 4 1 3 [a,b,c]
alt4Tupel a b c d  = Alt 5 1 4 [a,b,c,d]
altZero            = Alt 6 1 0 []
altOne             = Alt 6 2 0 []
altNpair a b       = Alt 6 3 2 [a,b]

-----------------------------------------------------------------------------------------------------------------------------------------------
-- actions (functions) chkAlts, chkAltPairs, chkAltCons, chkAlt3Tupel, chkAlt4Tupel prove wether case-alternatives correct and exhaustive
----------------------------------------------------------------------------------------------------------------------------------------------- 

chkAlts alts = case alts of
            [] -> error "case without alternatives";
            [Alt 3 1 b v exp] -> [Alt 3 1 b (showVar b v) exp];
            [Alt 4 1 b v exp] -> [Alt 4 1 b (showVar b v) exp];
            [Alt 5 1 b v exp] -> [Alt 5 1 b (showVar b v) exp];
            [Alt 1 1 b v1 exp1, Alt 1 2 d v2 exp2] -> [Alt 1 1 b (showVar b v1) exp1, Alt 1 2 d (showVar d v2) exp2];
            [Alt 2 1 b v1 exp1, Alt 2 2 d v2 exp2] -> [Alt 2 1 b (showVar b v1) exp1, Alt 2 2 d (showVar d v2) exp2];
            [Alt 1 2 b v1 exp1, Alt 1 1 d v2 exp2] -> [Alt 1 1 d (showVar d v2) exp2, Alt 1 2 b (showVar b v1) exp1];
            [Alt 2 2 b v1 exp1, Alt 2 1 d v2 exp2] -> [Alt 2 1 d (showVar d v2) exp2, Alt 2 2 b (showVar b v1) exp1];
            otherwise -> error "Exeption: invalid or too many alternatives in the case pattern"


showVar b  v = if (chkVar v == True) && (b == length (v)) then v 
               else error ("Exeption: missing variable in the pattern")

chkVar [] = True
chkVar (x:xs) = if any (x==) xs then error ("Exeption: repeated variables in the pattern")
                  else chkVar(xs)


chkAltPair :: Var -> Var -> Exp -> Alts
chkAltPair a b exp = if not (a == b) then Alt 3 1 2 [a,b] exp
		      	     else error ("Exeption: repeated variables in the pattern")
			      
chkAltCons :: Var -> Var -> Exp -> Alts
chkAltCons a b exp = if not (a == b) then Alt 1 2 2 [a,b] exp 
				     else error ("Exeption: repeated variables in the pattern")

chkAlt3Tupel :: Var -> Var -> Var -> Exp -> Alts
chkAlt3Tupel a b c exp = if [a,b,c] == nub [a,b,c] then Alt 4 1 3 [a,b,c] exp
                             else error ("Exeption: repeated variables in the pattern")

chkAlt4Tupel :: Var -> Var -> Var -> Var -> Exp -> Alts
chkAlt4Tupel a b c d exp = if [a,b,c,d] == nub [a,b,c,d] then Alt 5 1 4 [a,b,c,d] exp
                             else error ("Exeption: repeated variables in the pattern")

-------------------------------------------------------------------------------------------------------
-- the action chkValue proves wether the expression is a value
-------------------------------------------------------------------------------------------------------

chkVal :: Token -> Exp -> Exp
chkVal a exp = case exp of
               V var  -> exp;
               Constr a b c d -> exp;
               Lambda a b -> exp;
               CCell -> exp;
               CThread -> exp;
               CLazy -> exp;
               CUnit -> exp;
               CHandle -> exp;
               otherwise -> error ("Check the value input at the position " ++ show (getPos a))

-------------------------------------------------------------------------------------------------------------------------------------------------------
-- mkInt performes a binary encoding  of integers, all integers must have the following form: Constr 6 3 2 [x:xs]; 
-- Constr 6 3 2 is a special pair , where x is either True (if an integer is positive) or False (if negative)(as Constr 2 1 0 [] or Constr 2 2 0 []),
-- and xs is a list of Zeros and Ones (also as constructors), i.e. a binary code of an integer
------------------------------------------------------------------------------------------------------------------------------------------------------- 

mkInt :: Integer -> Exp
mkInt x = let xplus = if (x >= 0) then x else (-x); ifpos = if(x >= 0) then 1 else 2  in  Constr 6 3 2 [Constr 2 ifpos 0 [], mkInt2 xplus]
mkInt2 :: Integer -> Exp
mkInt2 0 = Constr 1 2 2 [Constr 6 1 0 [], Constr 1 1 0 []]
mkInt2 1 = Constr 1 2 2 [Constr 6 2 0 [], Constr 1 1 0 []]
mkInt2 x = Constr 1 2 2 [Constr 6 (if (mod x 2) == 1 then 2 else 1) 0 [], mkInt2 (div x 2)]

-------------------------------------------------------------------------------------------------------------------------------------------------------
-- parser is the main function in this module, as already described above,
-- it binds the lexer (lexer) and the generated parser (parse) together and returnes a parsed input as a code of type Proc
-------------------------------------------------------------------------------------------------------------------------------------------------------

parser :: String -> Proc
parse :: [Token] -> Proc
parser = (parse . lexer)

-------------------------------------------------------------------------------------------------------------------------------------------------------
-- some tests 
-------------------------------------------------------------------------------------------------------------------------------------------------------

test_p1 = parser " a <= b"
test_p2 = parser " <n> x( <n> y( y <h> x))"
test_p3 = parser " <n> x( x <= case x of {True -> False, False -> True})"
test_p4 = parser " <n> x( x <= case x of {[] -> False})" 
}


