module ReductionRules 

where
import Parser
import SemAnalysis
import Syntax
import Transformation
import Char
import Data.List
import Maybe
import Debug.Trace
import Decoding

--------------------------------------------------------------------------------------------------------------------------------------------------------
-- the module ReductionRules contains implementation of reduction rules for evaluation contexts E and F (see Chapter 4 of our master's thesis)
--------------------------------------------------------------------------------------------------------------------------------------------------------
-- the function eval1 performes evaluation of evaluation contexts E :: x <= ~E and F :: x <= ~F, 
-- i.e. evaluation of expressions ONLY in the FIRST eager thread; 
-- the arguments of eval1  are (p = the first process component), (ns = a list of new names (nu)), (ps = a list of the rest processe components)
-- and (frv = a list of fresh variables);
-- eval1 invokes the function eval2 which executes evaluation rules using pattern matching;
-- eval1 returns either Nothing, if the first component is not an eager thread or an identifier is bound to a variable,
-- or Just (Proc' (a list of new names (nu)) (evaluated thread and the rest components as a list) (a list of unused fresh variables))
---------------------------------------------------------------------------------------------------------------------------------------------------------

eval1 :: Comp -> [Var] -> [Comp] -> [Var] -> Maybe (Proc', [Var])
eval1 p ns ps frv =
        case p of 
           (Thread' x (V y)) -> Nothing
           (Thread' x e) -> let s = eval2 e ns ps frv 
                                in case s of 
                                Just (e', ns', ps', frv') -> Just (Proc' ns' ((Thread' x e'):ps'), frv');
                                Nothing -> Nothing
           otherwise -> Nothing


-----------------------------------------------------------------------------------------------------------------------------------------------------
-- the function eval2 executes reduction rules in ~E and ~F contexts, as described in Chapter 4 of our master's thesis;
-- the arguments are an expression, a list of new names (nu), a list of process components and a list of fresh variables,
-- eval2 returns either Nothing, if reduction was not successful, 
-- or Just (evaluated expression, a list of new names, a changed list of process components, and a list of unused fresh variables) 
-----------------------------------------------------------------------------------------------------------------------------------------------------

eval2 :: Exp -> [Var] -> [Comp] -> [Var] -> Maybe (Exp, [Var], [Comp], [Var])

-----------------------------------------------------------------------------------------------------------------------------------------------------
-- (BETA-CBV(ev)) is a beta-reduction rule, the function betaRed is invoked here for beta-reduction (see below)
-----------------------------------------------------------------------------------------------------------------------------------------------------

eval2 (App (Lambda x e1) e2) ns ps frv                  
                 | isValue e2 = case (betaRed e2 x e1 frv) of
                                 (e',frv') -> Just (e',ns,ps,frv')
                 | otherwise  = case eval2 e2 ns ps frv of
                                 Just (e',ns,ps,frv') -> eval2 (App (Lambda x e1) e') ns ps frv'
                                 Nothing -> Nothing

-----------------------------------------------------------------------------------------------------------------------------------------------------
-- (THREAD.NEW(ev)) spawns a new eager thread
-----------------------------------------------------------------------------------------------------------------------------------------------------

eval2 (App CThread e2) ns ps (z:frv)
                 | isValue e2 = Just ((V z), z:ns, ((Thread' z (App e2 (V z))):ps), frv)
                 | otherwise  = case eval2 e2 ns ps (z:frv) of
                                  Just (e',ns,ps,frv') -> Just (App CThread e',ns,ps,frv')
                                  Nothing -> Nothing  
                     
-----------------------------------------------------------------------------------------------------------------------------------------------------
-- (CELL.NEW(ev)) spawns a new reference cell
-----------------------------------------------------------------------------------------------------------------------------------------------------

eval2 (App CCell e2) ns ps (z:frv)
                 | isValue e2 = Just ((V z), z:ns, ((Cell' z e2):ps), frv)
                 | otherwise = case eval2 e2 ns ps (z:frv) of
                                 Just (e',ns,ps,frv') -> Just (App CCell e',ns,ps,frv')
                                 Nothing -> Nothing

----------------------------------------------------------------------------------------------------------------------------------------------------- 
-- (LAZY.NEW(ev)) spawns a new lazy thread
-----------------------------------------------------------------------------------------------------------------------------------------------------

eval2 (App CLazy e2) ns ps (z:frv)
                 | isValue e2 = Just ((V z), z:ns, ((Lazythread' z (App e2 (V z))):ps), frv)
                 | otherwise  = case eval2 e2 ns ps (z:frv) of
                                 Just (e',ns,ps,frv') -> Just (App CLazy e',ns,ps,frv')
                                 Nothing -> Nothing

----------------------------------------------------------------------------------------------------------------------------------------------------- 
-- (HANDLE.NEW(ev)) spawns a new handle
-----------------------------------------------------------------------------------------------------------------------------------------------------

eval2 (App CHandle e2) ns ps (z1:z2:frv)
                 | isValue e2 = Just ((App (App e2 (V z1)) (V z2)), z1:z2:ns, ((Handle' z2 z1):ps), frv)
                 | otherwise = case eval2 e2 ns ps (z1:z2:frv) of
                               Just (e',ns,ps,frv') -> Just (App CHandle e2,ns,ps,frv')
                               Nothing -> Nothing

------------------------------------------------------------------------------------------------------------------------------------------------------
-- (BUFF.NEW(ev)) spawns a new concurrent buffer
------------------------------------------------------------------------------------------------------------------------------------------------------

eval2 (App CBuffer e2) ns ps (z:frv)
                 | isValue e2 = Just ((V z), z:ns, ((Emptybuffer' z):ps), frv)
                 | otherwise  = case eval2 e2 ns ps (z:frv) of
                                 Just (e',ns,ps,frv') -> Just (App CBuffer e',ns,ps,frv')
                                 Nothing -> Nothing

------------------------------------------------------------------------------------------------------------------------------------------------------
-- the following evaluation is for reduction rules (FUT.DEREF(ev)), (LAZY.TRIGGER(ev)), (HANDLE.BIND(ev));
-- depending on a result (here s) of the function searchInProcs (see below), an appropriate rule will be executed;
-- (FUT.DEREF(ev)) reduction rule is used for dereferencing of a future value, thereby values with bound variables must be alpha-renamed (renameVal);
-- (LAZY.TRIGGER(ev)) is used for triggering of lazy threads, i.e. lazy threads become eager threads for the next computational step;
-- (HANDLE.BIND(ev)) binds a new future to a value
------------------------------------------------------------------------------------------------------------------------------------------------------

eval2 (App (V x) e2) ns ps (z:frv)
                 | isValue e2 = 
                   let s = searchInProcs x ps e2
                   in case s of
                        Just ((Thread' a b), e',ps') -> let (e'',zs') = renameVal e' (z:frv) []    -- (FUT.DEREF(ev))
                                                        in Just ((App e'' e2), ns, ps', zs');      
                        Just ((Lazythread' a b), e',ps') -> Just (App (V x) e2,ns,ps',z:frv);      -- (LAZY.TRIGGER(ev))
                        Just ((Handle' a b), e',ps') -> Just (e',ns,ps',(z:frv));                  -- (HANDLE.BIND(ev))
                        Just ((Usedhandle' a ), e',ps') -> error "gfd" -- Just (e',ns,ps',(z:frv));                  -- (HANDLE.BIND(ev))
                        Nothing -> Nothing
                 | otherwise = case eval2 e2 ns ps frv of
                                Just (e',ns,ps,frv') -> Just (App (V x) e',ns,ps,frv');
                                Nothing -> Nothing

-------------------------------------------------------------------------------------------------------------------------------------------------------
-- the following evaluation is for reduction rules (CELL.EXCH(ev)) and FUT.DEREF(ev))and (LAZY.TRIGGER(ev)) for exch-expression; 
-- depending on a result (here s) of the function searchInProcs (see below), an appropriate rule will be executed;
-- (CELL.EXCH(ev)) writes a new value into a reference cell and returns its prevoius value;
-- (FUT.DEREF(ev)) occurs if the first argument of exch-expression is a variable;
-- (LAZY.TRIGGER(ev)) occurs if the first argument of exch-expression is a variable;
-------------------------------------------------------------------------------------------------------------------------------------------------------

eval2 (Exch (V e1) e2) ns ps (z:frv)
                  | isValue e2 = 
                    let s = searchInProcs e1 ps e2
                    in case s of
                        Just ((Thread' a b), e',ps') -> let (e'', zs') = renameVal e' (z:frv) []           -- (FUT.DEREF(ev))
                                                        in Just ((Exch e'' e2), ns, ps', zs'); 
                        Just ((Cell' a b), e',ps') -> Just (e', ns, ps', (z:frv));                         -- (CELL.EXCH(ev))   
                        Just ((Lazythread' a b), e',ps') -> Just ((Exch (V e1) e2), ns, ps', (z:frv));     -- (LAZY.TRIGGER(ev))
                        Nothing -> Nothing

-------------------------------------------------------------------------------------------------------------------------------------------------------                                   
-- if one or both arguments of exch-expression are not values, they must be evaluated before applying
-------------------------------------------------------------------------------------------------------------------------------------------------------

eval2 (Exch e1 e2) ns ps frv
                  | not (isValue e1) = case eval2 e1 ns ps frv of
                    Just (e',ns,ps,frv') -> Just (Exch e' e2,ns,ps,frv')
                  | isValue e1 && not (isValue e2) = case eval2 e2 ns ps frv of
                    Just (e',ns,ps,frv') -> Just (Exch e1 e',ns,ps,frv')
                  | otherwise = Nothing


-------------------------------------------------------------------------------------------------------------------------------------------------------
-- the following evaluation is for reduction rules (BUFF.PUT(ev)) and FUT.DEREF(ev))and (LAZY.TRIGGER(ev)) for put-expression; 
-- depending on a result (here s) of the function searchInProcs (see below), an appropriate rule will be executed;
-- (BUFF.PUT(ev)) writes a new value into an empty buffer;
-- (FUT.DEREF(ev)) occurs if the first argument of put-expression is a variable;
-- (LAZY.TRIGGER(ev)) occurs if the first argument of put-expression is a variable;
-------------------------------------------------------------------------------------------------------------------------------------------------------

eval2 (Put (V e1) e2) ns ps (z:frv)
                  | isValue e2 = 
                    let s = searchInProcs e1 ps e2
                    in case s of
                        Just ((Thread' a b), e',ps') -> let (e'', zs') = renameVal e' (z:frv) []           -- (FUT.DEREF(ev))
                                                        in Just ((Put e'' e2), ns, ps', zs');                               
                        Just ((Emptybuffer' b), e',ps') -> Just (e',ns,ps',(z:frv));                       -- (BUFF.PUT(ev))
                        Just ((Lazythread' a b), e',ps') -> Just ((Put (V e1) e2), ns, ps', (z:frv));      -- (LAZY.TRIGGER(ev))
                        Just ((Buffer' a b), e',ps') -> error "Error: the put operation is blocked, the buffer is not empty";  
                        other -> Nothing
-------------------------------------------------------------------------------------------------------------------------------------------------------                                   
-- if one or both arguments of put-expression are not values, they must be evaluated before applying
-------------------------------------------------------------------------------------------------------------------------------------------------------
                                  
eval2 (Put e1 e2) ns ps frv
                  | not (isValue e1) = case eval2 e1 ns ps frv of
                    Just (e',ns,ps,frv') -> Just (Put e' e2,ns,ps,frv')
                  | isValue e1 && not (isValue e2) = case eval2 e2 ns ps frv of
                    Just (e',ns,ps,frv') -> Just (Put e1 e',ns,ps,frv')
                  | otherwise = Nothing

-------------------------------------------------------------------------------------------------------------------------------------------------------
-- the following evaluation is for reduction rules (BUFF.GET(ev)) and FUT.DEREF(ev))and (LAZY.TRIGGER(ev)) for applications with the constant get; 
-- depending on a result (here s) of the function searchInProcs (see below), an appropriate rule will be executed;
-- (BUFF.GET(ev))  takes a value from a non-empty buffer;
-- (FUT.DEREF(ev)) occurs if get is applied to a variable ;
-- (LAZY.TRIGGER(ev)) occurs if get is applied to a variable 
-------------------------------------------------------------------------------------------------------------------------------------------------------

eval2 (App CGet (V x)) ns ps (z:frv) =
                   let s = searchInProcs x ps (V x)
                   in case s of
                        Just ((Thread' a b), e',ps') -> let (e'', zs') = renameVal e' (z:frv) []           -- (FUT.DEREF(ev))
                                                        in Just ((App CGet e''), ns, ps', zs');
                        Just ((Buffer' a b), e',ps') -> Just (e',ns,ps',(z:frv));                          -- (BUFF.GET(ev))
                        Just ((Lazythread' a b), e',ps') -> Just ((App CGet (V x)), ns, ps', (z:frv));     -- (LAZY.TRIGGER(ev))
                        Just ((Emptybuffer' a ), e',ps') -> Nothing ; -- error "Error: get operation is blocked, the buffer is empty";  
                        Nothing -> Nothing
-------------------------------------------------------------------------------------------------------------------------------------------------------                                   
-- if one or both arguments an aplication are not values, they must be evaluated before applying
-------------------------------------------------------------------------------------------------------------------------------------------------------

eval2 (App e1 e2) ns ps frv   
                 | isValue e1 = case eval2 e2 ns ps frv of
                                Just (res,ns,ps,frv') -> Just (App e1 res,ns,ps,frv');
                                Nothing -> Nothing
                 
                 | otherwise  = case eval2 e1 ns ps frv of   
                                 Just (res,ns,ps,frv') -> Just (App res e2,ns,ps,frv')
                                 Nothing -> Nothing

------------------------------------------------------------------------------------------------------------------------------------------------------
-- (CASE.BETA(ev)) performs case-beta-reduction and invokes the function betaCase (see below)
-----------------------------  -----------------------------------------------------------------------------------------------------------------------

eval2 (Case e alts) ns ps frv = case e of 
                             (V x) -> let s = searchInProcs x ps e
                                        in case s of
                                        Just ((Thread' a b), e',ps') -> Just ((Case b alts), ns, ps', frv); 
                                        Nothing -> Nothing
                             (Constr a b c xs) -> case (betaCase (Case e alts) frv) of
                                                (e',frv') -> Just (e',ns,ps,frv');
                             otherwise -> case eval2 e ns ps frv of
                                         Just (e',ns,ps,frv') -> Just (Case e' alts,ns,ps,frv')
                                         Nothing -> Nothing

-------------------------------------------------------------------------------------------------------------------------------------------------------
-- in these cases no reduction is possible:
-------------------------------------------------------------------------------------------------------------------------------------------------------

eval2  (Lambda _ _) ns ps frv = Nothing
eval2  (Constr _ _ _ _) ns ps frv = Nothing
eval2  (V _) ns ps frv =  Nothing
eval2  (CUnit) _ _ _ = Nothing
          
--------------------------------------------------------------------------------------------------------------------------------------------------------
-- searchInProcs searches in a process for certain components and returns changed components and expressions 
-- in accordance with reduction rules, they are used for   
--------------------------------------------------------------------------------------------------------------------------------------------------------

searchInProcs :: Var -> [Comp] -> Exp -> Maybe (Comp, Exp, [Comp])
searchInProcs a x e = siP a x x e

siP a _ [] e2  = Nothing
siP a (x:xs) (y:ys) e2 = case x of
                           Thread' b e3 -> if (a==b) then
                                            if (isValue e3) then Just ((Thread' b e3), e3, (x:xs))
                                            else Nothing
                                           else let p = xs ++ [x] in (siP a p ys e2);
                           Lazythread' b e3 -> if a==b then  Just ((Lazythread' b e3), e3, ((Thread' a e3):xs))
                                                  else let p = xs ++ [x] in (siP a p ys e2);
                           Handle' b y -> if a==b then Just ((Handle' b y), CUnit, ((Thread' y e2):(Usedhandle' b):xs)) else let p = xs ++ [x] in (siP a p ys e2);
                           Cell' b v -> if (a==b)  && (isValue v) then Just ((Cell' b v), v, ((Cell' a e2):xs)) else let p = xs ++ [x] in (siP a p ys e2);
                           Emptybuffer' b -> if a==b then Just ((Emptybuffer' b), CUnit, ((Buffer' b e2):xs)) else let p = xs ++ [x] in (siP a p ys e2);
                           Buffer' b e3 -> if a==b then
                                             if (isValue e3) then Just ((Buffer' b e3), e3, ((Emptybuffer' b):xs))
                                             else Nothing
                                           else let p = xs ++ [x] in (siP a p ys e2);
                           Usedhandle' b  -> if a==b then Nothing -- error "Error: process applies to a used handle" ;
                                             else siP a (xs ++ [x]) ys e2 ;
                           otherwise -> error "searchInProcs a xs e2 ;"
-------------------------------------------------------------------------------------------------------------------------------------------------------
-- isValue tests wether an expression is a value
-------------------------------------------------------------------------------------------------------------------------------------------------------

isValue :: Exp -> Bool
isValue exp = case exp of
               V var  -> True;
               Constr a b c d -> True;
               Lambda a b -> True;
               CCell -> True;
               CThread -> True;
               CLazy -> True;
               CUnit -> True;
               CHandle -> True;
               CBuffer -> True;
               CGet -> True;
               otherwise -> False

--------------------------------------------------------------------------------------------------------------------------------------------------------
-- the function betaRed performs beta-reduction in expressions for the reduction rule (BETA.CBV(ev)),
-- it invokes subfunctions betaRedInConstr and betaRedInCase for beta-reductions in constructors and case-expressions, respectively
--------------------------------------------------------------------------------------------------------------------------------------------------------

betaRed :: Exp -> Var -> Exp -> [Var] -> (Exp, [Var])
betaRed t v (V x) y 
    | x == v    = let (t',zs) = renameVal t y [] in (t', zs) 
    | otherwise = ((V x), y)

betaRed t v (Lambda x s) y
    | x == v     = error "in betaRed :lambda v = x" -- ((Lambda x s), y)
    | otherwise   = let	(t1, t2) = (betaRed t v s y) in ((Lambda x t1), t2)


betaRed t v (App e1 e2) y = let (t1, t2) = (betaRed t v e1 y);
                                    (s1, s2) = (betaRed t v e2 t2)
                                in  ((App t1 s1), s2)

betaRed t v (Exch e1 e2) y = let (t1, t2) = (betaRed t v e1 y);
                                     (s1, s2) = (betaRed t v e2 t2)
                                 in  ((Exch t1 s1), s2)

betaRed t v (Put e1 e2) y = let (t1, t2) = (betaRed t v e1 y);
                                    (s1, s2) = (betaRed t v e2 t2)
                                in  ((Put t1 s1), s2)

betaRed t v CCell y = (CCell, y)
betaRed t v CThread y = (CThread, y)
betaRed t v CLazy y = (CLazy, y)
betaRed t v CHandle y = (CHandle, y)
betaRed t v CUnit y = (CUnit, y)
betaRed t v CBuffer y = (CBuffer, y)
betaRed t v CGet y = (CGet, y)

betaRed t v (Constr a b c args) y = let (t1, t2) = (betaRedInConstr t v args y) in ((Constr a b c t1), t2)

betaRed t v (Case e alts) y = let (t1, t2) = (betaRed t v e y );
			                   (s1, s2) = (betaRedInCase t v alts t2)
		                     in	 ((Case t1 s1), s2)

betaRedInConstr t v [] z = ([],z)
betaRedInConstr t v (y:ys) z = 
            let	(t1, t2) = (betaRed t v y z);  
                  (s1, s2) = (betaRedInConstr t v ys t2)	in (t1:s1, s2)

betaRedInCase t v [] z = ([],z)
betaRedInCase t v ((Alt y1 y2 k patvs y3):ys) z 
 | v `elem` patvs = error "v in patvars"
 | otherwise      =  let (t1, t2) = (betaRed t v y3 z);
                         (s1, s2) = (betaRedInCase t v ys t2)
                in      ((Alt y1 y2 k patvs t1):s1, s2)

-------------------------------------------------------------------------------------------------------------------------------------------------------
-- betaCase performs case-beta-reduction in case-expressions for the reduction rule (CASE.BETA(ev))
-------------------------------------------------------------------------------------------------------------------------------------------------------

betaCase :: Exp -> [Var] -> (Exp, [Var])
betaCase (Case (Constr a b c args) alts) frv  = 
                                       let (args',e') = match a b alts
                                       in (matchList (zip args args') e' frv)

match :: Int -> Int -> [Alts] -> ([Var], Exp)
match a b (alt@((Alt d e f args' e')):as)
       | a == d && b == e = (args',e')
       | otherwise = match a b as

matchList :: [(Exp, Var)] -> Exp -> [Var] -> (Exp, [Var])
matchList []         e  y = (e,y)
matchList ((t,v):xs) e  y = let (e',y') = betaRed t v e y 
                            in matchList xs e' y'

-------------------------------------------------------------------------------------------------------------------------------------------------------
-- some tests
-------------------------------------------------------------------------------------------------------------------------------------------------------

test_r1 = searchInProcs  "x" [Thread' "x" (V "s"), Thread' "d" (V "s")] (V "w")
test_r2 = eval1 (Thread' "s" (App (Lambda "s" (V "s")) (V "d"))) [] [] frv1
test_r3 = betaRed (V "d")  "s" (Lambda "a" (V "s")) frv1
test_r4 = betaRed (Lambda "d" (V "k"))  "s" (App (V "s") (V "s")) frv1
test_r5 = betaRed (Constr 2 1 0 [])  "s" (Lambda "g" (V "s")) frv1
test_r6 = betaCase (Case (Constr 1 2 2 [V "x",V "y"]) [Alt 1 1 0 [] (Constr 1 1 0 []),Alt 1 2 2 ["z","zs"] (V "zs")]) frv1 
t_r= (Case (V "e") [Alt 1 1 0 [] (Constr 1 1 0 []),Alt 1 2 2 ["z","zs"] (V "zs")])
test_r7 = eval2 t_r [] [Thread' "e" (Constr 1 2 2 [V "x",V "y"])] frv1

frv1 = ["fVar" ++show x|x<-[1..20]]