module SemAnalysis (isWellformed, renameBoundedProc, renameVal)

where
import Syntax
import Data.List
import Parser


--------------------------------------------------------------------------------------------------------------------------------------------
-- the module SemAnalysis performes the semantic analysis of a program of type Proc for the calculus lambda(fchb),
-- this semantic analysis consists of proving the well-formedness and variable bindings,
-- these bindings are lambda bindings in expressions and new operator (nu) bindings in processes
-------------------------------------------------------------------------------------------------------------------------------------------- 
-- the function isWellformed proves wether a program is well-formed, i.e. that there are no variables in processes used more than once:
-- it creates a list of all variables used in processes, then applies the function nub to this list, that removes duplicates from a list
-- and creates another list without duplicates. If these list are identic (True) than a program is well-formed
--------------------------------------------------------------------------------------------------------------------------------------------

isWellformed :: Proc -> Bool  
isWellformed s = if (varList s == nub (varList s)) then True 
                     else error ("is not wellformed, repeated variables:  " ++ show (varList s)) 
varList x = case x of 
                  (Parall a b)     -> (varList a) ++ (varList b);
                  (New a b)        -> if isWellformed b then [] else error ("is not wellformed, repeated variables:  " ++ show (varList b)); 
                  (Handle a b)     -> [a] ++ [b];
                  (Cell a b)       -> [a];
                  (Thread a b)     -> [a];
                  (Usedhandle a)   -> [a];
                  (Lazythread a b) -> [a];
                  (Buffer a b)     -> [a];
                  (Emptybuffer a)  -> [a]

------------------------------------------------------------------------------------------------------------------------------------------------
-- the function renameBoundedProc proves wether all variables in processes are bound , i.e. they are introduced by a new operator (nu),
-- and performes alpha-renaming of bound variables in processes using a list of fresh variables (freshVariables);
-- for checking of variable bindings within expressions the function renameBoundedExp will be called
------------------------------------------------------------------------------------------------------------------------------------------------


freshVariables = ["newVar" ++show x|x<-[1..500]]  
            
renameBoundedProc :: Proc -> Proc
renameBoundedProc x  = fst (renameNew x freshVariables [])

-----------------------------------------------------------------------------------------------------------------------------------------------
-- the function renameNew has three arguments: a process component, a list of fressh variables, and a list of variable pairs for comparing,
-- and returns a process with renamed bounded variables and a list of unused fresh variables
-----------------------------------------------------------------------------------------------------------------------------------------------

renameNew :: Proc -> [Var] -> [(Var,Var)] ->(Proc, [Var])
renameNew (Usedhandle a) y z = (Usedhandle (renameVar a z), y)
renameNew (Thread a b) y z = let (b',y') = renameBoundedExp b y z in (Thread (renameVar2 a z) b', y')
renameNew (Cell a b) y z = let (b',y') = renameBoundedExp b y z in (Cell (renameVar2 a z) b', y')
renameNew (Lazythread a b) y z = let (b',y')  = renameBoundedExp b y z in (Lazythread (renameVar2 a z) b', y')
renameNew (Handle a b) y z = (Handle (renameVar2 a z) (renameVar2 b z), y) 
renameNew (Buffer a b) y z = let (b',y') = renameBoundedExp b y z in (Buffer (renameVar2 a z) b',y')
renameNew (Emptybuffer a) y z = (Emptybuffer (renameVar2 a z), y)
renameNew (Parall p1 p2) y z = 
	let	(t1, t2) = (renameNew p1 y z);
		(s1, s2) = (renameNew p2 t2 z)
	in	((Parall t1 s1), s2)
renameNew (New x p) (y:ys) mappings = let (new_p, freshVariables') = renameNew p ys ((x,y):mappings)
                                                in (New y new_p, freshVariables') 


-------------------------------------------------------------------------------------------------------------------------------------------------
-- the subfunction renameVar catches free variables, it compares all occurrence of variables with a list of variable pairs (the second argument),
-- and reports if a variable is not bound
-------------------------------------------------------------------------------------------------------------------------------------------------

renameVar :: Var -> [(Var, Var)] -> Var
renameVar x [] = error ("Variable " ++ x ++ " is not bound") 
renameVar x ((z1,z2):zs) = if(z1 == x) then z2 else (renameVar x  zs)
                                     
-----------------------------------------------------------------------------------------------------------------------------------------------
-- the function renameBoundedExp proves wether all variables in expressions are bound with lambda bindings,
-- and analogical to renameBoundedProc performes alpha-renaming of bounded variables in expressions using a list of fresh variables
-----------------------------------------------------------------------------------------------------------------------------------------------

renameBoundedExp :: Exp -> [Var] -> [(Var, Var)] -> (Exp, [Var])
renameBoundedExp x fv mappings  = (renameLam x fv mappings) 

-----------------------------------------------------------------------------------------------------------------------------------------------
-- the subfunction renameLam is analogical to the function renameNew but has an expression as the first argument,
-- together with its subfunctions renameInConstr, renameInCase and renameVar it catches free variables 
-- and returns an expression with renamed bounded (with lambda bindings) variables and a list of unused fresh variables
-----------------------------------------------------------------------------------------------------------------------------------------------

renameLam :: Exp -> [Var] -> [(Var,Var)] ->(Exp, [Var])
renameLam (V x) y z = (V (renameVar x z), y)
renameLam (CCell) y z = (CCell, y)
renameLam (CThread) y z = (CThread, y)
renameLam (CHandle) y z = (CHandle, y)
renameLam (CLazy) y z = (CLazy, y)
renameLam (CUnit) y z = (CUnit, y)
renameLam (CBuffer) y z = (CBuffer, y)
renameLam (CGet) y z = (CGet, y)
renameLam (App x1 x2) y z = 
	let	(t1, t2) = (renameLam x1 y z);
		(s1, s2) = (renameLam x2 t2 z)
	in	((App t1 s1), s2)
renameLam (Exch x1 x2) y z = 
	let	(t1, t2) = (renameLam x1 y z);
		(s1, s2) = (renameLam x2 t2 z)
	in	((Exch t1 s1), s2)
renameLam (Put x1 x2) y z = 
	let	(t1, t2) = (renameLam x1 y z);
		(s1, s2) = (renameLam x2 t2 z)
	in	((Put t1 s1), s2)
renameLam (Lambda x1 x2) (y:ys) z =
		let	(t1, t2) = (renameLam x2 ys ((x1, y):z))
		in	((Lambda  y t1), t2)
renameLam (Constr x1 x2 x3 x4) y z =
		let	(t1, t2) = (renameInConstr z x4 y)
		in	((Constr x1 x2 x3 t1), t2)
renameLam (Case x1 x2) y z = 
		let	(t1, t2) = (renameLam x1 y z);
			(s1, s2) = (renameInCase z x2 t2)
		in	((Case t1 s1), s2)

renameInConstr :: [(Var,Var)] -> [Exp] -> [Var]->([Exp], [Var])
renameInConstr x [] z = ([], z)
renameInConstr x (y:ys) z = 
            let	(t1, t2) = (renameLam y z x);  
                  (s1, s2) = (renameInConstr x ys t2)	in (t1:s1, s2)

renameInConstrVal :: [(Var,Var)] -> [Exp] -> [Var]->([Exp], [Var])
renameInConstrVal x [] z = ([], z)
renameInConstrVal x (y:ys) z = 
            let (t1, t2) = (renameVal y z x);  
                  (s1, s2) = (renameInConstrVal x ys t2)   in (t1:s1, s2)

renameInCase :: [(Var,Var)]->[Alts]->[Var]->([Alts], [Var])
renameInCase x [] z = ([], z)
renameInCase x ((Alt y1 y2 0 [] y3):ys) z = 
		let	(t1, t2) = (renameLam y3 z x);
			(s1, s2) = (renameInCase x ys t2)
		in	((Alt y1 y2 0 [] t1):s1, s2)
renameInCase x ((Alt y1 y2 2 (v:w:[]) y3):ys) (z1:z2:zs) = 
		let	 (v1, w1, vnew, wnew) = (v, w, z1, z2); (t1, t2) = (renameLam y3 zs ((v1,z1):(w1,z2):x)); (s1, s2) = (renameInCase x ys t2)
		in	 ((Alt y1 y2 2 (vnew:wnew:[]) t1):s1, s2)
renameInCase x ((Alt y1 y2 3(u:v:w:[]) y3):ys) (z1:z2:z3:zs) = 
		let	 (u1, v1, w1, unew, vnew, wnew) = (u, v, w, z1, z2, z3); (t1, t2) = (renameLam y3 zs ((u1,z1):(v1,z2):(w1,z3):x)); (s1, s2) = (renameInCase x ys t2)
            in	 ((Alt y1 y2 3 (unew:vnew:wnew:[]) t1):s1, s2)
renameInCase x ((Alt y1 y2 4(q:u:v:w:[]) y3):ys) (z1:z2:z3:z4:zs) = 
		let	 (q1, u1, v1, w1, qnew, unew, vnew, wnew) = (q, u, v, w, z1, z2, z3, z4); (t1, t2) = (renameLam y3 zs ((q1,z1):(u1,z2):(v1,z3):(w1,z4):x)); (s1, s2) = (renameInCase x ys t2)
            in     ((Alt y1 y2 4 (qnew:unew:vnew:wnew:[]) t1):s1, s2)

----------------------------------------------------------------------------------------------------------------------------------------------------
-- the function renameVal is analog to the function renameLam, it differs only in the subfunction renameVar2,
-- renameVar2 allows free variables in expressions,
-- renameVal will be called in the module Evaluation for beta-reduction
----------------------------------------------------------------------------------------------------------------------------------------------------            

renameVal :: Exp -> [Var] -> [(Var,Var)] ->(Exp, [Var])
renameVal (V x) y z = (V (renameVar2 x z), y)
renameVal (CCell) y z = (CCell, y)
renameVal (CThread) y z = (CThread, y)
renameVal (CHandle) y z = (CHandle, y)
renameVal (CLazy) y z = (CLazy, y)
renameVal (CUnit) y z = (CUnit, y)
renameVal (CGet) y z = (CGet, y)
renameVal (App x1 x2) y z = 
        let     (t1, t2) = (renameVal x1 y z);
                (s1, s2) = (renameVal x2 t2 z)
        in      ((App t1 s1), s2)
renameVal (Exch x1 x2) y z = 
        let     (t1, t2) = (renameVal x1 y z);
                (s1, s2) = (renameVal x2 t2 z)
        in      ((Exch t1 s1), s2)
renameVal (Put x1 x2) y z = 
      let     (t1, t2) = (renameVal x1 y z);
              (s1, s2) = (renameVal x2 t2 z)
      in      ((Put t1 s1), s2)
renameVal (Lambda x1 x2) (y:ys) z =
                let     (t1, t2) = (renameVal x2 ys ((x1, y):z))
                in      ((Lambda  y t1), t2)
renameVal (Constr x1 x2 x3 x4) y z =
                let     (t1, t2) = (renameInConstrVal z x4 y)
                in      ((Constr x1 x2 x3 t1), t2)
renameVal (Case x1 x2) y z = 
                let     (t1, t2) = (renameVal x1 y z);
                        (s1, s2) = (renameInCaseVal z x2 t2)
                in      ((Case t1 s1), s2)

renameVar2 :: Var -> [(Var,Var)]-> Var
renameVar2 x  [] = x
renameVar2 x  ((z1,z2):zs) = if(z1 == x) then z2 else (renameVar2 x zs)

renameInCaseVal :: [(Var,Var)]->[Alts]->[Var]->([Alts], [Var])
renameInCaseVal x [] z = ([], z)
renameInCaseVal x ((Alt y1 y2 0 [] y3):ys) z = 
                let     (t1, t2) = (renameVal y3 z x);
                        (s1, s2) = (renameInCaseVal x ys t2)
                in      ((Alt y1 y2 0 [] t1):s1, s2)
renameInCaseVal x ((Alt y1 y2 2 (v:w:[]) y3):ys) (z1:z2:zs) = 
                let      (v1, w1, vnew, wnew) = (v, w, z1, z2); 
                         (t1, t2) = (renameVal y3 zs ((v1,z1):(w1,z2):x));
                         (s1, s2) = (renameInCaseVal x ys t2)
                in       ((Alt y1 y2 2 (vnew:wnew:[]) t1):s1, s2)
renameInCaseVal x ((Alt y1 y2 3(u:v:w:[]) y3):ys) (z1:z2:z3:zs) = 
                let      (u1, v1, w1, unew, vnew, wnew) = (u, v, w, z1, z2, z3); (t1, t2) = (renameVal y3 zs ((u1,z1):(v1,z2):(w1,z3):x));(s1, s2) = (renameInCaseVal x ys t2)
                   in    ((Alt y1 y2 3 (unew:vnew:wnew:[]) t1):s1, s2)
renameInCaseVal x ((Alt y1 y2 4(q:u:v:w:[]) y3):ys) (z1:z2:z3:z4:zs) = 
                let      (q1, u1, v1, w1, qnew, unew, vnew, wnew) = (q, u, v, w, z1, z2, z3, z4); (t1, t2) = (renameVal y3 zs ((q1,z1):(u1,z2):(v1,z3):(w1,z4):x));(s1, s2) = (renameInCaseVal x ys t2)
                in       ((Alt y1 y2 4 (qnew:unew:vnew:wnew:[]) t1):s1, s2)


------------------------------------------------------------------------------------------------------------------------------------------------------
-- some tests
------------------------------------------------------------------------------------------------------------------------------------------------------

test_s1 = isWellformed $ parser "<n> x (<n> y (x <= y | y <= buffer))"
test_s2 = isWellformed $ parser "<n> x (<n> y (<n> z (x <= y | y <= buffer)))"
test_s3 = renameBoundedProc $ parser "<n> x (<n> y (<n> z (x <= y | y <= buffer)))"
test_s4 = renameBoundedProc $ parser "<n> x (<n> y (<n> z (x <= y | y <= \\ s -> a)))"
test_s5 = renameBoundedProc $ parser "<n> x (<n> a (<n> b (<n> c (x <c> True | a <= c a | b <= c b | c <= \\z ->(thread (\\y -> exch(x,z)))))))"
