module Transformation (Proc' (..), Comp (..), transfToProc', transfToProcHtoB', transfToProcBtoH', procPrimeToProc)

where
import Syntax
import Data.List
import Parser
import SemAnalysis 
import qualified EncodingHtoB as EHB
import qualified EncodingBtoH as EBH
import Decoding

------------------------------------------------------------------------------------------------------------------------------------------------------
-- the module Transformation proviedes two transformations:
-- it simplifies the syntax of the code and transforms it into the new data structure data Proc' (function transfToProc'),
-- and transforms Proc' back to Proc (function procPrimeToProc);
------------------------------------------------------------------------------------------------------------------------------------------------------
-- data Proc' consists of a terminal Proc' and two lists: the first list is a list of all new operators (nu), 
-- and the second list is a list of all components, so the terminals New and Parall are to be eliminated  
------------------------------------------------------------------------------------------------------------------------------------------------------ 
-- to make the code more readable for users we create here show instanses and give the code its original view
------------------------------------------------------------------------------------------------------------------------------------------------------ 

data Proc' = Proc' [Var] [Comp]
             deriving (Eq)

instance Show Proc' where 
  show (Proc' vars comp) = concat $ (map (\v -> "<n> " ++ v ++ " (") vars) ++ (intersperse " | " $ map show comp) ++ (replicate  (length vars) ")")
                          
------------------------------------------------------------------------------------------------------------------------------------------------------

data Comp = Cell'  Var Exp
          | Thread' Var Exp
          | Handle' Var Var
          | Usedhandle' Var
          | Lazythread' Var Exp
          | Buffer'  Var Exp
          | Emptybuffer'  Var 
            deriving (Eq)

instance Show Comp where
 show (Cell' v e) = v ++ " <c> " ++ show1 e
 show (Thread' v e) = v ++ " <= " ++ show1 e
 show (Handle' v1 v2) = v1 ++ " <h> " ++ v2
 show (Usedhandle' v1 ) = v1 ++ " <h> -"
 show (Lazythread' v e) = v ++ " <s= " ++ show1 e
 show (Buffer' v e) = v ++ " <b> " ++ show1 e
 show (Emptybuffer' v ) = v ++ " <eb> -"
 
-----------------------------------------------------------------------------------------------------------------------------------------------------
-- the functions transfToProc' a program of type Proc into type Proc', 
-- but before this transformation occurs, a program will be proved whether it is well-formed and closed (i.e. no free variables occur);
-- transfToProc' makes the transformation for the calculus lambda(fchb), without any encoding
-----------------------------------------------------------------------------------------------------------------------------------------------------
-- the functions transfToProcHtoB' is like the function transfToProc',
-- but performes the trancformation for the calculus lambda(fcb), i.e. after translation (\(fch) -> \(fcb)
-----------------------------------------------------------------------------------------------------------------------------------------------------
-- the functions transfToProcBtoH' is like the function transfToProc',
-- but performes the trancformation for the calculus lambda(fch), i.e. after translation (\(fcb) -> \(fch)
-----------------------------------------------------------------------------------------------------------------------------------------------------

transfToProc' x = if isWellformed x == True 
                     then transf [] (renameBoundedProc x) else error "The process is not well-formed"

transfToProcHtoB' x = if isWellformed x == True 
                     then transf [] (EHB.renameBoundedProcHandleToBuffer x) else error "The process is not wellformed"

transfToProcBtoH' x = if isWellformed x == True 
                     then transf [] (EBH.renameBoundedProcBufferToHandle x) else error "The process is not wellformed"


transf v x = case x of
             (Cell a b) -> Proc' v [Cell' a b];
             (Thread a b) -> Proc' v [Thread' a b];
             (Handle a b) -> Proc' v [Handle' a b];
             (Lazythread a b) -> Proc' v [Lazythread' a b];
             (Usedhandle a) -> Proc' v [Usedhandle' a];
             (Buffer a b) -> Proc' v [Buffer' a b];
             (Emptybuffer a) -> Proc' v [Emptybuffer' a];
             (New a b) -> let  new_v = v ++ [a] in transf new_v b;   
             (Parall a b) -> let Proc' nuvars procs = addParall (transf [] a) (transf [] b)
                              in  Proc' (v++nuvars) procs

addParall (Proc' a b) (Proc' c d) = Proc' (a++c) (b++d)                  

-----------------------------------------------------------------------------------------------------------------------------------------------------
-- the function procPrimeToProc transforms Proc' back to Proc
-----------------------------------------------------------------------------------------------------------------------------------------------------

procPrimeToProc :: Proc' -> Proc
procPrimeToProc (Proc' vars comps) = foldr New (foldr1 Parall (map compToProc comps)) vars

compToProc :: Comp -> Proc
compToProc c = case c of
                 (Cell' v e) -> Cell v e
                 (Thread' v e) -> Thread v e
                 (Handle' v1 v2) -> Handle v1 v2
                 (Usedhandle' v) -> (Usedhandle v)
                 (Lazythread' v e) ->  (Lazythread v e)
                 (Buffer'  v e) -> (Buffer v e)
                 (Emptybuffer' v) -> (Emptybuffer v)

----------------------------------------------------------------------------------------------------------------------------------------------------
-- some tests
----------------------------------------------------------------------------------------------------------------------------------------------------

test_t1 =  transfToProc' $  parser $ "<n> x (<n> y (y <= (\\w -> put(x,False) ) (get x) | x <b> True ))"
test_t2 =  transfToProcBtoH' $  parser $ "<n> x (<n> y (y <= (\\w -> put(x,False) ) (get x) | x <b> True ))"
test_t3 =  transfToProcHtoB' $  parser $ "<n> x (<n> y (y <h> x | x <_> ))"
test_t4 =  transfToProcHtoB' $  parser $ "<n> x (<n> y (y <= handle unit))"
test_t5 =  procPrimeToProc $ transfToProc' $  parser $ "<n> x (<n> y (y <= (\\w -> put(x,False) ) (get x) | x <b> True ))"
